/** \file    consensus.c
 *  \brief   Consensus procedure
 *  \author  Takahisa Toda
 *  \see     Figure 3.1 at page 131 of "Yves Crama and Peter L. Hammer, Boolean Functions: Theory, Algorithms, and Applications, Encyclopedia of Mathematics and Its Applications 142, Cambridge, 2011".
 */

#include <stdio.h>
#include <assert.h>
#include "consensus.h"

static bool isabsorbed(veci* t1, veci* t2, int *w);
static bool compute_consensus(veci* t1, veci* t2, veci* t3, int *w);

/* \brief obtain a free veci from fv.
 * \param fv vector that holds free vecis
 * \return pointer to a free veci
 */
static inline veci *getveci(vecp *fv)
{
    veci *new;
    if (vecp_size(fv) > 0) {
        new = vecp_begin(fv)[vecp_size(fv)-1];
        vecp_resize(fv, vecp_size(fv)-1);
    } else {
        new = (veci*)malloc(sizeof(veci));
        veci_new(new);
    }
    veci_resize(new, 0);

    return new;
}


/* \brief retrieve an unnecessary veci.
 * \param fv vector that holds free vecis
 * \param v unnecessary veci
 */
static inline void retrieveveci(vecp *fv, veci *v)
{
    vecp_push(fv, v);
}


/* \brief delete all free vecis and delete fv.
 * \param fv vector that holds free vecis
 */
static inline void freeveci_delete(vecp *fv)
{
    for (int i = 0; i < vecp_size(fv); i++) {
        veci_delete(vecp_begin(fv)[i]);
        free(vecp_begin(fv)[i]);
    }

    vecp_delete(fv);
}


static inline void insertdebug(int n, int *w) // for debug
{
#ifdef DEBUG
    for (int k = 0; k < n; k++)
        assert(w[k] == l_Undef);
#endif

}


static inline void print_unchecked_checked(vecp *unchecked, vecp *checked) // for debug
{
#ifdef DEBUG
        printf("------------------------------------------------------------------------\n");
        printf("unchecked:\n");
        for (int i = 0;i < vecp_size(unchecked); i++)
            print_term(vecp_begin(unchecked)[i]);
        printf("checked:\n");
        for (int i = 0;i < vecp_size(checked); i++)
            print_term(vecp_begin(checked)[i]);
        printf("\n");
#endif
}

static inline void print_progress(int *old,  dnf_t *unchecked, dnf_t *g)
{
#ifdef PROGRESS
        int new = (int)(100 * (double)vecp_size(g)/(vecp_size(unchecked)+vecp_size(g)));
        if (abs(new - *old) > 2) {
            int i = 0;
            printf("\r[");
            for (; i <= new; i += 2)
                printf("=");
            printf(">");
            for (; i <= 100; i += 2)
                printf(" ");
            printf("]");
            printf(" %d%%", new);
            fflush(stdout);
            *old = new;
        }
#endif
}

void consensus(dnf_t *f, dnf_t *g)
{
    vecp_resize(g, 0);

    vecp freeveci;
    vecp_new(&freeveci);
    vecp_resize(&freeveci, 0);

    dnf_t unchecked;
    vecp_new(&unchecked);

    int max = -1;
    for (int i = 0; i < vecp_size(f); i++) {
        veci *t   = vecp_begin(f)[i];
        int *lits = veci_begin(t);
        int n     = veci_size(t);
        veci *u   = getveci(&freeveci);
        for (int j = 0; j < n; j++) {
           max = max < lit_var(lits[j]) ? lit_var(lits[j]): max; 
           veci_push(u, lits[j]);
        }
        vecp_push(&unchecked, u);
    }
    assert(max >= 0);
    max++; // increment by one since variable index starts from 0.

    int* w = (int*)malloc(sizeof(int)*max);
    for (int i = 0; i < max; i++)
        w[i] = l_Undef;

    int progress = 0;
#ifdef PROGRESS
            printf("\r[=>");
            for (int i = 0; i <= 100; i += 2)
                printf(" ");
            printf("]");
            printf(" 0%%");
            fflush(stdout);
#endif
    while (vecp_size(&unchecked) > 0) {
        veci *t = vecp_pop(&unchecked);

        print_progress(&progress, &unchecked, g);

        // check if there exists a term s among checked terms such that s is absorbed by t.
        bool flag1 = false;
        for (int i = 0; i < vecp_size(g);) {
            veci *s = vecp_begin(g)[i]; 
            if (isabsorbed(s, t, w)) {
                vecp_begin(g)[i] = vecp_begin(g)[vecp_size(g)-1];
                vecp_resize(g, vecp_size(g)-1);
            } else {
                i++;
            }

            if (isabsorbed(t, s, w)) {
                flag1 = true;
                break;
            }
        }

        if (flag1) {
            veci_delete(t);
            free(t);
            continue;
        }

        // check if there exist a term s among checked terms such that s and t have a consensus u and u is not absorbed by another term.
        for (int i = 0; i < vecp_size(g); i++) {
            veci *s = vecp_begin(g)[i]; 
            veci *u = getveci(&freeveci);
            bool flag2 = compute_consensus(s,t,u,w);
            if (flag2) {
                assert(veci_size(u) > 0);
                for (int j = 0; flag2 && j < vecp_size(g); j++) {
                    if (isabsorbed(u, vecp_begin(g)[j], w))
                        flag2 = false;
                }
                for (int j = 0; flag2 && j < vecp_size(&unchecked); j++) {
                    if (isabsorbed(u, vecp_begin(&unchecked)[j], w))
                        flag2 = false;
                }
            }

            if (flag2)
                vecp_push(&unchecked, u); 
            else
                retrieveveci(&freeveci, u);
        }

        vecp_push(g, t);
    }

    freeveci_delete(&freeveci); 
    dnf_delete(&unchecked);

#ifdef PROGRESS
    printf("\n");
#endif
}


/** \brief decide if t1 is absorbed by t2, that is, t1 or t2 = t2
 *  \param t1 term
 *  \param t2 term
 *  \param w  working space of length equal to t1 (and also to t2)
 *  \return true if absorbed; false, otherwise.
 */
static bool isabsorbed(veci* t1, veci* t2, int *w)
{
    if (t1 == t2)
        return true;

    int *lits1 = veci_begin(t1);
    int *lits2 = veci_begin(t2);

#ifdef DEBUG
    printf("t1:\t");
    print_term(t1);
    printf("t2:\t");
    print_term(t2);
#endif

    for (int k = 0; k < veci_size(t1); k++)
        w[lit_var(lits1[k])] = lit_sign(lits1[k])? l_False: l_True; // set

    bool res = true;
    for (int k = 0; k < veci_size(t2); k++) {
        if (w[lit_var(lits2[k])] != (lit_sign(lits2[k])? l_False: l_True)) {
            res = false;
            break;
        }
    }

    for (int k = 0; k < veci_size(t1); k++)
        w[lit_var(lits1[k])] = l_Undef;   // reset

#ifdef DEBUG
    if(res)
        printf("absorbed\n");
    else
        printf("not absorbed\n");
    printf("\n");
#endif

    return res;
}


/*  \brief compute consensus of t1 and t2 if exists.
 *  \param t1 term
 *  \param t2 term
 *  \param t3 term, in which result is stored.
 *  \param w  working space of length equal to t1 (and also to t2)
 *  \return true if exists; false, otherwise
 */
static bool compute_consensus(veci* t1, veci* t2, veci* t3, int *w)
{
    int *lits1 = veci_begin(t1);
    int *lits2 = veci_begin(t2);

#ifdef DEBUG
    printf("t1:\t");
    print_term(t1);
    printf("t2:\t");
    print_term(t2);
#endif

    veci_resize(t3, 0);

    // find variable x on which consensus is computed.
    for (int k = 0; k < veci_size(t1); k++)
        w[lit_var(lits1[k])] = lit_sign(lits1[k])? l_False: l_True; // set

    int x;
    int cnt = 0; // the number of oposite variables
    for (int k = 0; k < veci_size(t2); k++) {
        if (w[lit_var(lits2[k])] == (lit_sign(lits2[k])? l_True: l_False)) {
            cnt++;
            if (cnt >= 2) // this means no consensus exists.
                break;
            x = lit_var(lits2[k]);
            lit tmp = *veci_begin(t2);
            veci_begin(t2)[0] = veci_begin(t2)[k];
            veci_begin(t2)[k] = tmp;
        }
    }

    if (cnt == 1) { // x is found
        assert(veci_size(t1) > 0);
        assert(veci_size(t2) > 0);
        for (int k = 0; k < veci_size(t1); k++) {
            if (lits1[k] == lit_neg(lits2[0])) {
                lit tmp = *veci_begin(t1);
                veci_begin(t1)[0] = veci_begin(t1)[k];
                veci_begin(t1)[k] = tmp;
                break;
            }
        }
        assert(*veci_begin(t1) == lit_neg(*veci_begin(t2)));

        for (int k = 1; k < veci_size(t1); k++) {
            veci_push(t3, veci_begin(t1)[k]);
        }
        for (int k = 1; k < veci_size(t2); k++) {
            if (w[lit_var(lits2[k])] != l_Undef) {
               assert(w[lit_var(lits2[k])] == (lit_sign(lits2[k])? l_False: l_True)); 
            } else {
                veci_push(t3, veci_begin(t2)[k]);
            }
        }
    }

    for (int k = 0; k < veci_size(t1); k++)
        w[lit_var(lits1[k])] = l_Undef;   // reset

#ifdef DEBUG
    if (cnt == 1) {
        printf("t3:\t");
        print_term(t3);
    }
    printf("\n");
#endif

    return cnt == 1;
}
