/**************************************************************************************************
MiniSat -- Copyright (c) 2005, Niklas Sorensson
http://www.cs.chalmers.se/Cs/Research/FormalMethods/MiniSat/

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/
// Modified to compile with MS Visual Studio 6.0 by Alan Mishchenko
// Modified to implement consensus procedure by Takahisa Toda

#include "basic_types.h"
#include "consensus.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
//#include <unistd.h>
//#include <signal.h>
//#include <zlib.h>
//#include <sys/time.h>
//#include <sys/resource.h>

//=================================================================================================
// Helpers:


// Reads an input stream to end-of-file and returns the result as a 'char*' terminated by '\0'
// (dynamic allocation in case 'in' is standard input).
//
char* readFile(FILE *  in)
{
    char*   data = malloc(65536);
    int     cap  = 65536;
    int     size = 0;

    while (!feof(in)){
        if (size == cap){
            cap *= 2;
            data = realloc(data, cap); }
        size += fread(&data[size], 1, 65536, in);
    }
    data = realloc(data, size+1);
    data[size] = '\0';

    return data;
}

//static inline double cpuTime(void) {
//    struct rusage ru;
//    getrusage(RUSAGE_SELF, &ru);
//    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; }


//=================================================================================================
// DIMACS Parser:


static inline void skipWhitespace(char** in) {
    while ((**in >= 9 && **in <= 13) || **in == 32)
        (*in)++; }

static inline void skipLine(char** in) {
    for (;;){
        if (**in == 0) return;
        if (**in == '\n') { (*in)++; return; }
        (*in)++; } }

static inline int parseInt(char** in) {
    int     val = 0;
    int    _neg = 0;
    skipWhitespace(in);
    if      (**in == '-') _neg = 1, (*in)++;
    else if (**in == '+') (*in)++;
    if (**in < '0' || **in > '9') fprintf(stderr, "PARSE ERROR! Unexpected char: %c\n", **in), exit(1);
    while (**in >= '0' && **in <= '9')
        val = val*10 + (**in - '0'),
        (*in)++;
    return _neg ? -val : val; }

static int readClause(char** in, veci* lits) {
    int parsed_lit, var;
    veci_resize(lits,0);
    int max = -1;

    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        veci_push(lits, (parsed_lit > 0 ? toLit(var) : lit_neg(toLit(var))));
        max = max < var+1 ? var+1 : max;
    }

    return max;
}

static lbool parse_DIMACS_main(char* in, dnf_t *f) {
    for (;;){
        skipWhitespace(&in);
        if (*in == 0)
            break;
        else if (*in == 'c' || *in == 'p')
            skipLine(&in);
        else{
            veci* lits = (veci*)malloc(sizeof(veci));
            veci_new(lits);
            readClause(&in, lits);
            if (veci_size(lits) > 0)
                vecp_push(f,lits);
            else
                {veci_delete(lits); free(lits); return l_False;} 
        }
    }

    return l_True;
}


// Inserts problem into solver. Returns FALSE upon immediate conflict.
//
static lbool parse_DIMACS(FILE * in, dnf_t *f) {
    char* text = readFile(in);
    lbool ret  = parse_DIMACS_main(text, f);
    free(text);
    return ret; }


//=================================================================================================

//solver* slv;
//static void SIGINT_handler(int signum) {
//    printf("\n"); printf("*** INTERRUPTED ***\n");
//    printStats(&slv->stats, cpuTime());
//    printf("\n"); printf("*** INTERRUPTED ***\n");
//    exit(0); }


//=================================================================================================

static inline void PRINT_USAGE(char *p)
{
    fprintf(stderr, "Usage:\t%s input-file [output-file]\n", (p));
}

int main(int argc, char** argv)
{
    lbool   st;
    FILE *  in;
    FILE *  out;
    int     clk = clock();

    char *infile  = NULL;
    char *outfile = NULL;

  /*** RECEIVE INPUTS ***/  
  for(int i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch (argv[i][1]){
      case '?': case 'h': default:
        PRINT_USAGE(argv[0]); return  0;
      }   
    } else {
      if(infile == NULL)        {infile  = argv[i];}
      else if(outfile == NULL)  {outfile = argv[i];}
      else                      {PRINT_USAGE(argv[0]); return  0;}
    }   
  }
  if(infile == NULL) {PRINT_USAGE(argv[0]); return  0;}

    in = fopen(infile, "rb");
    if (in == NULL)
        fprintf(stderr, "ERROR! Could not open file: %s\n", argc == 1 ? "<stdin>" : infile),
        exit(1);
    if(outfile != NULL) {
      out = fopen(outfile, "wb");
      if (out == NULL)
        fprintf(stderr, "ERROR! Could not open file: %s\n", argc == 1 ? "<stdin>" : outfile),
        exit(1);
    } else {
      out = NULL;
    }

    dnf_t f, g;
    vecp_new(&f);
    vecp_new(&g);

    st = parse_DIMACS(in, &f);
    fclose(in);
    if (st == l_False) {
        dnf_delete(&f);
        exit(20);
    }

    consensus(&f, &g);
    double Time    = (float)(clock()-clk)/(float)(CLOCKS_PER_SEC);
    printf("input             : %s\n", infile);
    printf("CPU time          : %12.2f sec\n", Time);
    printf("clauses(in)       : %12.0d\n", vecp_size(&f));
    printf("clauses(out)      : %12.0d\n", vecp_size(&g));
    fflush(stdout);
   
    if (out != NULL) { 
        dnf_sort(&g);
        const int m = vecp_size(&g);
        for (int i = 0; i < m; i++)
            fprint_term(out, vecp_begin(&g)[i]);
    }

    dnf_delete(&f);
    dnf_delete(&g);
    if (out != NULL)
        fclose(out);

    return 0;
}
